/******************************************************************************
*                                                                             *
* This is a part of the LsxCommand LiteStep module Source code.               *
*                                                                             *
* Copyright (C) 1999-2000 Visigoth (Shaheen Gandhi in real life)              *
* Based on limpid's lscommand.                                                *
* Updated by blkhawk, ilmcuts, jesus_mjjg, rabidcow.                          *
* Look at the documentation for more informations.                            *
*                                                                             *
* This program is free software; you can redistribute it and/or modify        *
* it under the terms of the GNU General Public License as published by        *
* the Free Software Foundation; either version 2 of the License, or           *
* (at your option) any later version.                                         *
*                                                                             *
* This program is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of              *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
* GNU General Public License for more details.                                *
*                                                                             *
* You should have received a copy of the GNU General Public License           *
* along with this program; if not, write to the Free Software                 *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA   *
*                                                                             *
******************************************************************************/

/******************************************************************************
*                  history.c - Provides History Management                    *
*                                   *  *  *  *                                *
* Rewritten circa 2004-04-03                                                  *
******************************************************************************/

#ifndef LSXCOMMANDCLOCK_EXPORTS

#include <windows.h>
#include <stdio.h>
#include <limits.h> /* for CHAR_BIT */ 
#include "lsxcommand.h"
#include "list.h"

#include "ModuleLog.h"

struct list *lHistory = NULL;
HMENU hHistoryMenu = NULL;

extern char szModuleIniPath[MAX_PATH];

/* log(2)/log(10) = ~.301, rounding up is hard, so just add 1 */ 
#define BITS_TO_DIGITS(BITS) ((BITS)*301/1000 + 1)

/***********************************************************
* void History_Read()                                      *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct CommandSettings *cs;                         *
*      A valid CommandSettings structure.                  *
*                         *  *  *  *                       *
* Reads the history list from the file.                    *
***********************************************************/
static void History_Read(struct CommandSettings *cs)
{
  char entry[4096];
  unsigned long i;
  char name[BITS_TO_DIGITS(sizeof(i)*CHAR_BIT) + 1];

  if (lHistory)
  {
    for(i=0; i < lHistory->nMaxEntries; ++i)
	  {
		  _ultoa(i, name, 10);
		  GetPrivateProfileString("LSXCOMMAND", name, "", entry, sizeof(entry)/sizeof(*entry), szModuleIniPath);
		  if (*entry)
        List_QueueInsert(lHistory, entry);
		  else
        /* do not skip blank entries - see History_Write */ 
			  break;
	  }
  }
}

/***********************************************************
* void History_Write()                                     *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct CommandSettings *cs;                         *
*      A valid CommandSettings structure.                  *
*                         *  *  *  *                       *
* Writes the history list to the file.                     *
***********************************************************/
static void History_Write(struct CommandSettings *cs)
{
  const struct list_node *pNode;
  unsigned long i = 0;
  char name[BITS_TO_DIGITS(sizeof(i)*CHAR_BIT) + 1];

  if (lHistory && lHistory->pHead)
  {
    pNode = lHistory->pHead->pPrev;
    do
    {
		  _ultoa(i++, name, 10);
		  WritePrivateProfileString("LSXCOMMAND", name, pNode->sText, szModuleIniPath);
      pNode = pNode->pPrev;
    }
    while (pNode != lHistory->pHead->pPrev);
  }

  /* write a blank one to terminate the list */ 
	_ultoa(i, name, 10);
	WritePrivateProfileString("LSXCOMMAND", name, NULL, szModuleIniPath);

  /* actually, that should remove the entry, but   */ 
  /* it works out the same when you try to read it */ 
}

/***********************************************************
* void History_Load()                                      *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct CommandSettings *cs;                         *
*      A valid CommandSettings structure.                  *
*                         *  *  *  *                       *
* Loads the history list from the file and creates the     *
* initial context submenu.                                 *
***********************************************************/
void History_Load(struct CommandSettings *cs)
{
  REQUIRE(READABLE(cs));

  lHistory = List_Create(cs->MaxHistoryEntries);
  History_Read(cs);
  History_CreateMenu(cs);
}

/***********************************************************
* void History_Unload()                                    *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct CommandSettings *cs;                         *
*      A valid CommandSettings structure.                  *
*                         *  *  *  *                       *
* Saves the history list back to the file and releases the *
* resources allocated in History_Load.                     *
***********************************************************/
void History_Unload(struct CommandSettings *cs)
{
  REQUIRE(READABLE(cs));

  lHistory = List_Destroy(lHistory);
  DestroyMenu(hHistoryMenu);
  hHistoryMenu = NULL;
}

/***********************************************************
* void History_CreateMenu()                                *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct CommandSettings *cs;                         *
*      A valid CommandSettings structure.                  *
*                         *  *  *  *                       *
* Creates or updates the history submenu.                  *
***********************************************************/
void History_CreateMenu(struct CommandSettings *cs)
{
  struct list_node *pNode;
  int iInsert;
  UINT id = HMI_USER_HISTORY;

  REQUIRE(READABLE(cs));

  if (!hHistoryMenu)
    hHistoryMenu = CreatePopupMenu();

  if (hHistoryMenu)
  {
    /* We don't want to destroy and recreate the menu, */ 
    /* because then you'd have to update the entry in  */ 
    /* the main context menu, which would mean finding */ 
    /* out what position it went in again.  Instead,   */ 
    /* we remove all items from the existing submenu   */ 
    /* and refill it.  This also guarantees that it'll */ 
    /* be up to date no matter what state it was in.   */ 
    while (RemoveMenu(hHistoryMenu, 0, MF_BYPOSITION))
      ;
    /* 0 = top, -1 = bottom of menu */ 
    iInsert = (cs->NewestHistoryItemsOnTop) ? 0 : -1;
    pNode = lHistory->pHead;
    if (pNode)
    {
      do
      {
        ASSERT(READABLE(pNode));

        InsertMenu(hHistoryMenu, iInsert,
          MF_STRING|MF_ENABLED, ++id, pNode->sText);
        pNode = pNode->pNext;
      }
      while (pNode != lHistory->pHead);
    }
  }
}

/***********************************************************
* void History_Clear()                                     *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct CommandSettings *cs;                         *
*      A valid CommandSettings structure.                  *
*                         *  *  *  *                       *
* Removes all entries from the history list in memory and  *
* stored in the file.  This does NOT update the submenu.   *
***********************************************************/
void History_Clear(struct CommandSettings *cs)
{
  unsigned nMaxEntries = lHistory->nMaxEntries;
  List_Destroy(lHistory);
  lHistory = List_Create(nMaxEntries);
  History_Write(cs);
}

/***********************************************************
* void History_Browse()                                    *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - int Direction;                                      *
*      Where to move in the history list.                  *
*                         *  *  *  *                       *
* Move the browse position in the history list.            *
***********************************************************/
void History_Browse(int Direction)
{
  List_Browse(lHistory, Direction, cs->WrapHistory, (lHistory->nMaxEntries*HISTORY_PAGE_FRACTION));
}

/***********************************************************
* const char *History_GetCurrent()                         *
*                         *  *  *  *                       *
* Retrieves the text in the history entry that the browse  *
* pointer is pointing at, or NULL.                         *
***********************************************************/
const char *History_GetCurrent(void)
{
  REQUIRE(READABLE(lHistory));

  if (!lHistory->pBrowse)
    return NULL;
  else
    return lHistory->pBrowse->sText;
}

/***********************************************************
* void History_Add()                                       *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - const char *line;                                   *
*      The line to add to history.                         *
*                         *  *  *  *                       *
* Adds a line to the history list, maintaining the maximum *
* queue length and removing duplicates when appropriate.   *
***********************************************************/
extern struct CommandSettings *cs;
void History_Add(const char *line)
{
  REQUIRE(READABLE(line));

  List_QueueInsertUnique(lHistory, Parse_SkipSpace(line));
  History_Browse(BROWSE_HOME);
  History_Write(cs);
}

/***********************************************************
* unsigned History_Complete()                              *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - char *sBuffer;                                      *
*      Buffer containing the prefix and to which the       *
*      completed string will be written to.                *
*    - unsigned cchBuffer;                                 *
*      Size of sBuffer in chars.                           *
*    - unsigned cchPrefix;                                 *
*      Length of the existing prefix in chars.  This       *
*      should not include any terminating nul.             *
*                         *  *  *  *                       *
* Finds the first item in the history list matching the    *
* given prefix and copies the tail end to the buffer.      *
* Returns the final length of the string in the buffer.    *
***********************************************************/
unsigned History_Complete(char *sBuffer, unsigned cchBuffer, unsigned cchPrefix)
{
  unsigned cchAdjust = 0;
  struct list_node *pNode;

  REQUIRE(WRITEABLEN(sBuffer, cchBuffer));
  REQUIRE(cchPrefix < cchBuffer);

  /* history entries have spaces trimmed, so trim before matching */ 
  while (cchAdjust < cchPrefix && isspace((unsigned char)*sBuffer))
  {
    ++sBuffer;
    ++cchAdjust;

    ASSERT(READABLE(sBuffer));
  }

  cchPrefix -= cchAdjust;
  cchBuffer -= cchAdjust;

  ASSERT(WRITEABLEN(sBuffer, cchBuffer));
  ASSERT(cchPrefix < cchBuffer);

  /* was it all spaces? */ 
  if (cchPrefix == 0)
    return cchPrefix + cchAdjust;

  pNode = List_FindPrefixNode(lHistory, sBuffer, cchPrefix);

  /* nothing found? */ 
  if (!pNode)
    return cchPrefix + cchAdjust;

  ASSERT(READABLE(pNode));

  /* got it, copy it to their buffer */ 
  strncpy(sBuffer+cchPrefix, pNode->sText+cchPrefix, cchBuffer-cchPrefix-1);
  sBuffer[cchBuffer-1] = 0; /* make sure it's nul terminated */ 

  return strlen(sBuffer) + cchAdjust;
}

/*


 
*/

struct list *lFiles = NULL;

/***********************************************************
* void Files_Enumerate()                                   *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - const char *sPattern;                               *
*      The pattern to find files matching.                 *
*                         *  *  *  *                       *
* Add files matching sPattern to the file list the file    *
* list will be created if it doesn't exist                 *
***********************************************************/
static void Files_Enumerate(const char *sPattern)
{
  HANDLE hFind;
  WIN32_FIND_DATA FindData;

  hFind = FindFirstFile(sPattern, &FindData);
  if (hFind != INVALID_HANDLE_VALUE)
  {
    if (!lFiles)
      lFiles = List_Create(0);

    do
    {
      /* don't know if these will actually occur,  */ 
      /* but the old code checks them, so why not? */ 

      if (!*FindData.cFileName)
        continue;
      if (strcmp(FindData.cFileName, ".") == 0)
        continue;
      if (strcmp(FindData.cFileName, "..") == 0)
        continue;

      List_QueueInsert(lFiles, FindData.cFileName);
    }
    while (FindNextFile(hFind, &FindData));
    FindClose(hFind);

    if (lFiles->pHead)
      lFiles->pBrowse = lFiles->pHead;
    else
      lFiles = List_Destroy(lFiles);
  }
}

/***********************************************************
* unsigned Files_Complete()                                *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - char *sBuffer;                                      *
*      Buffer containing the prefix and to which the       *
*      completed string will be written to.                *
*    - unsigned cchBuffer;                                 *
*      Size of sBuffer in chars.                           *
*    - unsigned cchPrefix;                                 *
*      Length of the existing prefix in chars.  This       *
*      should not include any terminating nul.             *
*    - int Direction;                                      *
*      Which direction to move in the file list.           *
*                         *  *  *  *                       *
* Copies the tail of the selected item from the file list  *
* into the given buffer, creating the file list if it does *
* not exist.  Returns the length of the resulting string.  *
***********************************************************/
unsigned Files_Complete(char *sBuffer, unsigned cchBuffer, unsigned cchPrefix, int Direction)
{
  char *sToken;
  sToken = Parse_FindLastFileToken(sBuffer, cchPrefix);
  if (!sToken) return cchPrefix;

  /* first time: fill the autocomplete list */ 
  if (!lFiles)
  {
    if (cchPrefix + 2 >= cchBuffer) return cchPrefix;
    sBuffer[cchPrefix] = '*';
    sBuffer[cchPrefix+1] = 0;
    Files_Enumerate(sToken);
    sBuffer[cchPrefix] = 0;
    if (!lFiles)
      return cchPrefix;
  }

  /* select the next one and use it */ 
  List_Browse(lFiles, Direction, TRUE, 10);

  strncpy(sBuffer+cchPrefix, lFiles->pBrowse->sText+cchPrefix,
    cchBuffer-cchPrefix);

  return strlen(sBuffer);
}

/***********************************************************
* void Files_Close()                                       *
*                         *  *  *  *                       *
* Frees the file list.                                     *
***********************************************************/
void Files_Close(void)
{
  lFiles = List_Destroy(lFiles);
}

#endif //LSXCOMMANDCLOCK_EXPORTS
